/*
Here is a program that produces solutions to the xquinto puzzle,
recently posted by Dev Joneja.  The method it uses is very
inefficient, but it works well enough to produce solutions for any
size puzzle you are likely to do by hand.  I use it to produce X
backgrounds; many of the 30 x 30 solutions are interestingly
asymmetric.

,
Eamonn
*/
/* xq.c - solve quinto puzzle */

/* By Eamonn McManus <emcmanus@gr.osf.org>, April 1992.
   This program is not copyrighted and may be used freely.

   Solve the quinto puzzle, as encapsulated in the program xquinto
   posted to alt.sources by Dev Joneja <dj7@cunixf.cc.columbia.edu>.
   In this puzzle, you start with a rectangular array of grey
   squares and the object is to convert it into an array of white
   squares.  At each move you reverse the colour of any square and
   all the adjoining (not diagonally) squares.

   The final colour of a square is white if the number of adjoining
   squares including itself that have been selected is odd.  Hence
   for an m x n array we can construct m x n simultaneous equations
   in XOR.  We solve these equations by Gaussian elimination,
   probably far from the most efficient technique.

   There may be more than one solution, manifested by redundant
   equations.  (For instance if a solution is not completely
   symmetric then its rotations and reflections will also appear
   as solutions.)  By default we pick one such solution at random,
   by setting all uneliminated variables to 0.  If given the -a
   flag, we print all possible solutions, without eliminating
   rotations and reflections.  If given the -c flag we just say
   how many solutions there are.

   The output is a rectangle of characters with `X' for squares
   that should be selected and `.' for those that should not.  This
   can be converted by the X program atobm (see bitmap(1)) into an
   X bitmap that can be used for pretty backgrounds and the like.

   For an n x n square, the program uses O(n^4) memory and runs in
   O(n^4) time.  Much more efficient implementations are possible;
   the present one will typically not go much further than n = 40.
 */

#include <stdio.h>

extern void *calloc(int nobj, int size);

#define xmalloc(ptr, size) \
    if (((ptr) = calloc((size), sizeof *(ptr))) == NULL) { \
	perror("calloc"); exit(1); \
    } else

struct square {
    char value, elim, known;
    char *eqn;
    int eqncount;
};

#if DEBUG
static void showeqn(char *eqp, int depth, int width);
static void showeqns(struct square *grid, int depth, int width);
#endif

#define coord(i, j) ((i) * width + j)
#define showcoord(n) printf("(%d,%d)", n / width, n % width)


int main(int argc, char **argv) {
    int depth, width, size;
    int i, j;
    struct square *grid;
    char *eqns;
    char count = 0, all = 0;
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	if (!argv[i][1])
	    goto Usage;
	for (j = 1; argv[i][j]; j++) {
	    switch (argv[i][j]) {
	    case 'c':
		count = 1; break;
	    case 'a':
		all = 1; break;
	    default:
		goto Usage;
	    }
	}
    }
    if (i + 2 != argc) {
Usage:
	fprintf(stderr, "Usage: xq [-ac] depth width\n");
	exit(1);
    }
    depth = atoi(argv[i]);
    width = atoi(argv[i + 1]);
    if (depth <= 0 || width <= 0)
	goto Usage;
    size = depth * width;
    xmalloc(grid, size);
    xmalloc(eqns, size * (size + 1));

    /* Construct the equations.  eqp[n] is set iff square n is a
       term in the equation.  eqp[size] is the RHS.  */
    for (i = 0; i < depth; i++) {
	for (j = 0; j < width; j++) {
	    char *eqp = eqns + coord(i, j) * (size + 1);
	    grid[coord(i, j)].eqn = eqp;
	    eqp[size] = 1;
	    eqp[coord(i, j)] = 1;
	    if (i > 0)
		eqp[coord(i - 1, j)] = 1;
	    if (j > 0)
		eqp[coord(i, j - 1)] = 1;
	    if (i + 1 < depth)
		eqp[coord(i + 1, j)] = 1;
	    if (j + 1 < width)
		eqp[coord(i, j + 1)] = 1;
	}
    }

    /* Eliminate.  For each equation, find a nonzero term and eliminate that
       variable from subsequent equations by substitution.  Substitution is
       equivalent to XORing the equations together.  */
    for (i = 0; i < size; i++) {
	char *eqp = grid[i].eqn;
	for (j = 0; j < size; j++) {
	    if (eqp[j]) {
		int k;
#if DEBUG
		showcoord(i); printf(":"); showeqn(eqp, depth, width);
		printf(": elim "); showcoord(j); putchar('\n');
#endif
		grid[j].elim = 1;
		for (k = i + 1; k < size; k++) {
		    char *eqk = grid[k].eqn;
		    if (eqk[j]) {
			int l;
			for (l = 0; l < size + 1; l++)
			    eqk[l] ^= eqp[l];
		    }
		}
#if DEBUG
		showeqns(grid, depth, width);
#endif
		break;
	    }
	}
#if DEBUG
	if (j == size) {
	    showcoord(i);
	    printf(": nothing to elim\n");
	}
#endif
    }

    /* Now print the solution(s).  The last equation ought to have
       at most one unknown variable, the second-last at most two,
       etc.  Work backwards deriving a new known value from each
       equation.

       If there are multiple solutions, some of the equations will
       have been reduced to 0=0.  In this case, all of the uneliminated
       variables are free.  Set them to known values (initially 0)
       and proceed.  If -a is given, these variables are clocked
       in binary through all possible values in this loop.  */
    while (1) {
	j = 0;
	for (i = 0; i < size; i++) {
	    grid[i].known = !grid[i].elim;
	    j += grid[i].known;
	}
	if (count) {
	    printf("%d degree%s of freedom => ", j, j == 1 ? "" : "s");
	    if (j > 30)
		printf("2^%d solutions\n", j);
	    else printf("%d solution%s\n", 1 << j, j == 0 ? "" : "s");
	    return 0;
	}
	/* Derive the values.  */
	for (i = size - 1; i >= 0; i--) {
	    char *eqp = grid[i].eqn;
	    int unknown = -1;
	    char value = 0;
	    /* Find the unknown term in the equation. */
	    for (j = 0; j < size; j++) {
		if (eqp[j]) {
		    if (grid[j].known)
			value ^= grid[j].value;
		    else if (unknown < 0)
			unknown = j;
		    else {			/* Should not happen.  */
			showcoord(i); printf(": two unknowns: ");
			showcoord(unknown); printf(" and "); showcoord(j);
			putchar('\n');
			return 1;
		    }
		}
	    }
	    if (unknown >= 0) {
		grid[unknown].known = 1;
		grid[unknown].value = value ^ eqp[size];
#if DEBUG
		showcoord(unknown);
		printf(" = %d\n", grid[unknown].value);
#endif
	    } else if (value != eqp[size]) {	/* 0 = 1.  */
		showcoord(i);
		printf("inconsistent equations -- no solution\n");
		return 1;
	    } /* else redundant equation */
	}
	/* Print the solution.  */
	for (i = 0; i < depth; i++) {
	    for (j = 0; j < width; j++) {
		int x = coord(i, j);
		if (!grid[x].known) {		/* Should not happen. */
		    putchar('\n'); showcoord(x);
		    printf(": not known (should be!)\n");
		    return 1;
		}
		putchar(grid[x].value ? 'X' : '.');
	    }
	    putchar('\n');
	}
	if (!all)
	    break;
	/* Count up one on the binary clock of free variables.  */
	for (i = 0; i < size; i++) {
	    if (!grid[i].elim) {
		if (grid[i].value)
		    grid[i].value = 0;
		else {
		    grid[i].value = 1;
		    break;
		}
	    }
	}
	if (i == size)
	    break;
	putchar('\n');
    }
    return 0;
}


#if DEBUG
void showeqns(struct square *grid, int depth, int width) {
    int i, j;
    for (i = 0; i < depth; i++) {
	for (j = 0; j < width; j++) {
	    printf("(%d,%d):", i, j);
	    showeqn(grid[coord(i, j)].eqn, depth, width);
	    putchar('\n');
	}
    }
}


void showeqn(char *eqp, int depth, int width) {
    int i, j;
    if (eqp[depth * width])
	printf(" not");
    for (i = 0; i < depth; i++)
	for (j = 0; j < width; j++)
	    if (eqp[coord(i, j)])
		printf(" (%d,%d)", i, j);
}
#endif


